import sys
from plumbum import SshMachine, PuttyMachine

_WIN32 = sys.platform.startswith('win')


def RemoteMachine(*pargs, **kwargs):
    '''
    Remote machine constructor function.  Forwards all arguments on to the
    appropriate constructor for this platform.  On windows this is
    ``plumbum.PuttyMachine`` and on other platforms `plumbum.SshMachine`
    '''
    if _WIN32:
        return PuttyMachine(*pargs, **kwargs)
    else:
        return SshMachine(*pargs, **kwargs)





import re
import functools
from collections import namedtuple
from ConfigParser import ConfigParser
from contextlib import contextmanager
from plumbum import ProcessExecutionError

__all__ = ['Repo']


class Repo(object):
    '''
    This class provides an abstraction around running commands on a mercurial
    repository.  It can be used against either a local or remote repository
    depending on the ``machine`` parameter to the constructor.
    '''

    SummaryInfo = namedtuple('SummaryInfo', ['commit', 'mq'])
    CommitChangeInfo = namedtuple(
            'CommitChangeInfo',
            ['modified', 'unknown']
            )
    MqAppliedInfo = namedtuple('MqAppliedInfo', ['applied', 'unapplied'])

    # Template Parameter for hg log-style commands
    HgTemplateParam = '{node}\\t{desc|firstline}\\n'

    # Contains details of a changeset
    ChangesetInfo = namedtuple('ChangesetInfo', ['hash', 'desc'])
    ChangesetInfoRegexp = re.compile(r'^(?P<hash>\w+)\t(?P<desc>.*)$')

    def __init__(self, machine, remote=None):
        '''
        :param machine:     The plumbum machine object to use
                            (can be a local machine or remote machine)
        :param remote:      The name of the remote repo to be used by
                            push, pull and other operations.
        '''
        self.machine = machine
        self.hg = self.machine['hg']
        self.remote = remote
        self._currentRev = self._branch = None
        self.prevLevel = None

    @contextmanager
    def CleanMq(self):
        '''
        Returns a context manager that keeps the mq repository clean
        for it's lifetime
        '''
        revertTo = self.lastAppliedPatch
        self.PopPatch()
        yield
        if revertTo:
            self.PushPatch(revertTo)

    def _CleanMq(func):
        '''
        Decorator that ensures a function is always run with no patches applied
        Should only be applied on Repo methods

        :params func:   The function to decorate
        '''
        @functools.wraps(func)
        def InnerFunc(self, *pargs):
            with self.CleanMq():
                return func(self, *pargs)
        return InnerFunc

    @property
    def summary(self):
        '''
        Gets info from hg summary

        :return:    A :class:`SummaryInfo` containing :class:`CommitChangeInfo`
                    & :class:`MqAppliedInfo`
        '''
        commitData = Repo.CommitChangeInfo(0, 0)
        mqData = Repo.MqAppliedInfo(0, 0)
        commitRegexp = re.compile(
                r'^commit:\s+((\d+) modified(, (\d+) unknown)?)?'
                )
        mqRegexp = re.compile(r'^mq:\s+((\d+) applied, (\d+) unapplied)?')
        lines = self.hg('summary').splitlines()
        for line in lines:
            match = commitRegexp.search( line )
            if match:
                commitData = Repo.CommitChangeInfo(*match.group( 2, 4 ))
            match = mqRegexp.search( line )
            if match:
                mqData = Repo.MqAppliedInfo(*match.group(2, 3))
        return Repo.SummaryInfo(commitData, mqData)

    @property
    def currentRev(self):
        '''
        Gets the current revision
        This property is cached, so it may be out of date

        :returns:   A string containing the current revision hash
        '''
        if not self._currentRev:
            self._CheckCurrentRev()
        return self._currentRev

    @property
    def branch(self):
        '''
        Gets the current branch
        This property is cached, so it may be out of date

        :returns:   A string containing the current branch name
        '''
        if not self._branch:
            self._CheckCurrentRev()
        return self._branch

    @_CleanMq
    def _CheckCurrentRev( self ):
        ''' Gets the current revision and branch and stores it '''

        revMatch = re.search(
            r'^(\w{12})\+?\s+(.*)\s*$',
            self.hg('id', '-i', '-b')
            )
        if revMatch is None:
            raise Exception("Could not get current revision using hg id")
        self._currentRev, self._branch = revMatch.group(1, 2)

    def _RunListCommand(self, command, headerLines=0):
        '''
        Runs an hg command that gets a list (outgoing, incoming etc.)

        :param command:     The plumbum command object to run
        :param headerLines: The number of lines to chop off the top of
                            the output
        '''
        try:
            lines = command().splitlines()
            if headerLines == 0:
                return lines
            if len(lines) < headerLines:
                raise Exception("Unexpected number of lines from hg command")
            return lines[headerLines:]
        except ProcessExecutionError as e:
            if e.retcode != 1:
                # retcode of 1 just means there's nothing in the list,
                # so ignore it
                raise
        return []

    def _GetChangesetInfoList(self, *pargs, **kwargs):
        '''
        Utility function that calls _RunListCommand and filters the results
        through ChangesetInfoRegexp

        :returns:   A list of :class:`ChangesetInfo`
        '''
        lines = self._RunListCommand(*pargs, **kwargs)
        matches = (self.ChangesetInfoRegexp.match(line) for line in lines)
        return [self.ChangesetInfo(**match.groupdict()) for match in matches]

    @property
    @_CleanMq
    def outgoings(self):
        '''
        Gets the outgoing changesets to `self.remote`

        :returns:   A list containing :class:`ChangesetInfo` that represent
                    the current outgoing changesets
        '''
        assert self.remote
        return self._GetChangesetInfoList(
                self.hg[ 'outgoing', '-b', self.branch, '-r', self.currentRev,
                         '--template', self.HgTemplateParam, self.remote
                         ],
                headerLines=2
                )

    @property
    @_CleanMq
    def incomings(self):
        '''
        Gets the incoming changesets from `self.remote`

        :returns:   A list containing :class:`ChangesetInfo` that represent
                    the current incoming changesets
        '''
        assert self.remote
        return self._GetChangesetInfoList(
                self.hg[ 'incoming', '-b', self.branch,
                         '--template', self.HgTemplateParam, self.remote
                         ],
                headerLines=2
                )

    @property
    def lastAppliedPatch(self):
        '''
        Gets the last applied mq patch (if there is one)

        :returns: A single mq patch name (or None)
        '''
        try:
            patches = self._RunListCommand(self.hg['qapplied'])
            if patches:
                return patches[-1]
        except ProcessExecutionError as e:
            if e.retcode != 255:
                # 255 means mq is probably disabled
                raise
        return None

    @_CleanMq
    def PushToRemote(self):
        ''' Pushes to the remote repository at `self.remote`'''
        assert self.remote
        self.hg('push', '-b', self.branch, '-r', self.currentRev, self.remote)

    def PushMqToRemote(self):
        ''' Pushes the mq repo to the remote at `self.remote` '''
        assert self.remote
        try:
            self.hg('push', '--mq', self.remote)
        except ProcessExecutionError as e:
            if e.retcode != 1:
                #1 just means there's no outgoings
                raise

    def PopPatch(self, patch=None):
        '''
        Pops mq patch(es)

        :param patch:   Name of the patch to pop to.
                        If None, all will be popped
        '''
        # Check there are some patches applied
        level = self.lastAppliedPatch
        if level:
            if patch is None:
                patch = '-a'
            self.hg('qpop', patch)

    def PushPatch(self, patch=None):
        '''
        Pushes mq patch(es)

        :param patch:   Name of the patch to push to.
                        If None, all will be pushed
        '''
        if patch is None:
            patch = '-a'
        self.hg('qpush', patch)

    @_CleanMq
    def Strip(self, changesets):
        '''
        Strips changesets from this repository

        :param changesets:  A list of :class:`ChangesetInfo`
                            representing the changesets to strip
        '''
        self.hg('strip', *[cs.hash for cs in changesets])

    @_CleanMq
    def Update(self, changeset):
        '''
        Updates to a specific changeset

        :param changeset:   A changeset hash string, or
                            :class:`ChangesetInfo` representing
                            the changeset to update to
        '''
        if isinstance(changeset, self.ChangesetInfo):
            changeset = changeset.hash
        self.hg('update', changeset)

    def UpdateMq(self):
        '''
        Updates the mq repository to tip
        '''
        self.hg('update', '--mq')

    def RefreshMq(self):
        '''
        Refreshes the current mq patch
        '''
        self.hg('qrefresh')

    def CommitMq(self, msg=None):
        '''
        Commits the mq repository

        :param msg:     An optional commit message
        '''
        if not msg:
            msg = 'synchg-commit'
        try:
            self.hg('commit', '--mq', '-m', msg)
        except ProcessExecutionError as e:
            if e.retcode != 1:
                #1 just means there's no changes
                raise

    @_CleanMq
    def Clone(self, destination, createRemote=True):
        '''
        Clones the repository to a different location

        :param destination:     The destination clone path
        :param createRemote:    If set a remote will be created in the local
                                hgrc with the name the class was initialised
                                with.
        '''
        remoteName = self.remote if createRemote else None
        self._DoClone(destination, remoteName)
        patches = self.machine.cwd / '.hg' / 'patches'
        if patches.exists():
            # Clone mq repository
            # TODO: Possibly do a sanity check and call hg init --mq if needed
            with self.machine.cwd(patches):
                mqdest = destination + '/.hg/patches'
                self._DoClone(mqdest, remoteName)

    def _DoClone(self, destination, remoteName):
        '''
        Actually performs a clone operation

        :param destination:     The destination clone path
        :param remoteName:      The name of the remote to create (if any)
        '''
        self.hg('clone', '.', destination)
        if remoteName:
            config = self.machine.cwd / '.hg' / 'hgrc'
            hgconfig = ConfigParser()
            hgconfig.readfp(config.open())
            hgconfig.set('paths', remoteName, destination)
            hgconfig.write(config.open('w'))



'''
This module provides the actual syncing functionality for SyncHg.  It's
functions can be called by imported and called by other libraries if they wish
to make use of SyncHg functionality.
'''

import plumbum
from remote import RemoteMachine
from repo import Repo
from utils import yn


class AbortException(Exception):
    '''
    An exception that's thrown when a user chooses to abort.  This should be
    caught and ignored at the start of the program to allow users to abort at
    prompts
    '''
    pass


class SyncError(Exception):
    '''
    An exception that's thrown when a non-exceptional error occurs.  This
    exception is usually accompanied by an error message and should probably
    be caught and the backtrace suppressed.
    '''
    pass


def SyncRemote(host, name, localpath, remote_root):
    '''
    Syncs a remote repository.  This function should be called to kick off a
    sync

    :param host:        The hostname of the remote repository
    :param name:        The name of the project that is being synced.
                        This parameter will be appended to the remote_root
                        to find the remote repository.
    :param localpath:   A plumbum path to the local repository
    :param remote_root: The path to the parent directory of the
                        remote repository
    '''
    print "Sync {0} -> {1}".format(name, host)
    with RemoteMachine(host) as remote:
        with plumbum.local.cwd(localpath):
            local = Repo(plumbum.local, host)
            rpath = remote.cwd / remote_root / name
            if not rpath.exists():
                print "Remote repository can't be found."
                if yn('Do you want to create a clone?'):
                    local.Clone('ssh://{0}/{1}/{2}'.format(
                        host, remote_root, name
                        ))
                else:
                    raise AbortException
            with remote.cwd(rpath):
                _DoSync(local, Repo(remote))


def _DoSync(local, remote):
    '''
    Function that actually handles the syncing after everything
    has been set up

    :param local:   The local repository
    :param remote:  The remote repository
    '''
    # First, check the state of each repository
    if remote.summary.commit.modified:
        # Changes might be lost on remote...
        raise SyncError('Remote repository has uncommitted changes')

    lsummary = local.summary
    if lsummary.commit.modified:
        print "Local repository has uncommitted changes."
        if lsummary.mq.applied:
            # We can't push/pop patches to check remote is
            # in sync if we've got local changes, so prompt to refresh.
            if yn('Do you want to refresh the current patch?'):
                local.RefreshMq()
            else:
                print "Ok.  Please run again after dealing with changes."
                raise AbortException
        else:
            # If we're not doing an mq sync, we can happily ignore
            # these changes, but probably want to make sure that's
            # what the user wants...
            if not yn('Do you want to ignore these changes?'):
                print "Ok.  Please run again after dealing with changes."
                raise AbortException

    # Pop any patches on the remote before we begin
    remote.PopPatch()

    with local.CleanMq():
        if local.outgoings:
            incomings = local.incomings
            if incomings:
                # Don't want to be creating new remote heads when we push
                print "Changesets will be stripped from remote:"
                for hash, desc in incomings:
                    if len(desc) > 50:
                        desc = desc[:47] + '...'
                    print "  {0}  {1}".format(hash[:6], desc)
                if not yn('Do you want to continue?'):
                    raise AbortException()
                remote.Strip(incomings)
            print "Pushing to remote"
            local.PushToRemote()

    print "Updating remote"
    remote.Update(local.currentRev)

    appliedPatch = local.lastAppliedPatch
    if appliedPatch:
        print "Syncing mq repos"
        local.CommitMq()
        local.PushMqToRemote()
        print "Updating remote mq repo"
        remote.UpdateMq()
        remote.PushPatch(appliedPatch)

    print "Ok!"



import os
import synchg
from ConfigParser import ConfigParser, Error as ConfigParserError
from plumbum import cli, local
from clint import resources
from .sync import SyncRemote, AbortException, SyncError


class SyncHg(cli.Application):
    PROGNAME = 'SyncHg'
    VERSION = synchg.__version__
    DESCRIPTION = 'Syncs a remote mercurial repository'

    ConfigFileName = 'config.ini'

    def __init__(self, *pargs):
        super(SyncHg, self).__init__(*pargs)
        self.config = ConfigParser()

    def _get_config(self, in_do_config=False):
        '''
        Reads the configuration
        '''
        resources.init('obmarg', 'synchg')
        self.config = ConfigParser()
        contents = resources.user.read(self.ConfigFileName)
        if not contents:
            if not in_do_config:
                print "Could not find config file"
                self.do_config()
        else:
            self.config.readfp(resources.user.open(self.ConfigFileName))

    name = cli.SwitchAttr(
            ['n', '--name'],
            help='The directory name of the repository on the remote. '
                 'Uses the local directory name by default'
            )

    @cli.switch(['-c', '--config'])
    def do_config(self):
        '''
        Runs through the initial configuration
        '''
        print "Running initial configuration"
        self._get_config(True)
        srcdir = default = ''
        if not self.config.has_section('config'):
            self.config.add_section('config')
        try:
            default = self.config.get('config', 'hgroot')
        except ConfigParserError:
            pass
        while not srcdir:
            srcdir = raw_input(
                    "Remote source directory? [{0}] ".format(default)
                    )
        self.config.set('config', 'hgroot', srcdir)
        if not os.path.exists(resources.user.path):
            os.makedirs(resources.user.path)
        self.config.write(resources.user.open(self.ConfigFileName, 'w'))
        pass

    def main(self, remote_host, local_path=None):
        self._get_config()
        if local_path:
            local_path = local.cwd / local_path
        else:
            local_path = local.cwd

        if not self.name:
            self.name = local_path.basename

        SyncRemote(remote_host, self.name, local_path,
                   self.config.get('config', 'hgroot'))


def run():
    try:
        SyncHg.run()
    except AbortException:
        pass
    except SyncError as e:
        # TODO: Colour would be nice here..
        print "Error: {0}".format(e)



# -*- coding: utf8 -*-

"""
This file has been copied from clint on github, until an actual release contains
this functionality
"""

from re import match, I

def yn(prompt, default='y', batch=False):
    # A sanity check against default value
    # If not y/n then y is assumed
    if default not in ['y', 'n']:
        default = 'y'

    # Let's build the prompt
    choicebox = '[Y/n]' if default == 'y' else '[y/N]'
    prompt = prompt + ' ' + choicebox + ' '

    # If input is not a yes/no variant or empty
    # keep asking
    while True:
        # If batch option is True then auto reply
        # with default input
        if not batch:
            input = raw_input(prompt).strip()
        else:
            print prompt
            input = ''

        # If input is empty default choice is assumed
        # so we return True
        if input == '':
            return True

        # Given 'yes' as input if default choice is y
        # then return True, False otherwise
        if match('y(?:es)?', input, I):
            return True if default == 'y' else False

        # Given 'no' as input if default choice is n
        # then return True, False otherwise
        elif match('n(?:o)?', input, I):
            return True if default == 'n' else False




try:
    from distribute_setup import use_setuptools
    use_setuptools()
except ImportError:
    pass

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='SyncHg',
    version='0.9.7',
    url='http://synchg.readthedocs.org/',
    license='BSD',
    author='Graeme Coupar',
    author_email='grambo@grambo.me.uk',
    description='A simple script & library to handle syncing remote '
                'mercuial repositories',
    long_description=open('README.rst').read(),
    # if you would be using a package instead use packages instead
    # of py_modules:
    packages=['synchg'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'plumbum>=1.0.1',
        'clint>=0.3.1'
    ],
    entry_points={
        'console_scripts': [
            'synchg = synchg.script:run'
            ]
        },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities',
        'Topic :: Software Development :: Version Control',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
